import warnings
from datetime import datetime
from enum import Enum
from typing import Any, Callable, Dict, Iterable, List, Optional, Union

from pykechain.exceptions import IllegalArgumentError
from pykechain.utils import Empty, empty, is_url, is_uuid, parse_datetime

iter_types = (list, tuple, set)


def check_type(value: Optional[Any], cls: Any, key: str) -> Optional[Any]:
    """Validate any input to be an instance a specific class.

    :param value: value to check against
    :param cls: check to see if the value is of type class
    :param key: name of the object to display in the error message
    """
    if value is not None and value is not empty:
        if not isinstance(value, cls):
            if isinstance(cls, iter_types):
                types = ", ".join(c.__name__ for c in cls)
            else:
                types = cls.__name__
            raise IllegalArgumentError(
                '`{}` should be of type {}, "{}" ({}) is not.'.format(
                    key, types, value, type(value)
                )
            )
    return value


def check_client(value: "Client") -> "Client":
    """Validate the input being a KE-chain Client.

    :param value: the client to test
    """
    from pykechain import Client

    return check_type(value, Client, "client")


def check_uuid(uuid: Optional[str], key: Optional[str] = "pk") -> Optional[str]:
    """Validate the UUID input to be a correct UUID."""
    if uuid is not None and uuid is not empty:
        if not is_uuid(uuid):
            raise IllegalArgumentError(
                f'`{key}` must be a valid UUID, "{uuid}" ({type(uuid)}) is not.'
            )
    return uuid


def check_text(text: Optional[str], key: str) -> Optional[str]:
    """Validate text input to be a string."""
    if text is not None and text is not empty:
        if not isinstance(text, str):
            raise IllegalArgumentError(
                f'`{key}` should be a string, "{text}" ({type(text)}) is not.'
            )
    return text


def check_url(url: Optional[str], key: str = "url") -> Optional[str]:
    """Validate text input to be a valid format URL."""
    url = check_text(text=url, key=key)
    if url is not None and url is not empty:
        if not is_url(url):
            raise IllegalArgumentError(
                f'`{key}` should be a valid URL, "{url}" ({type(url)}) is not.'
            )
    return url


def check_list_of_text(
    list_of_text: Optional[Iterable[str]], key: str, unique: bool = False
) -> Optional[List[str]]:
    """Validate iterable input to be a list/tuple/set of strings."""
    if list_of_text is not None and list_of_text is not empty:
        if not isinstance(list_of_text, iter_types) or not all(
            isinstance(t, str) for t in list_of_text
        ):
            raise IllegalArgumentError(
                '`{}` should be a list, tuple or set of strings, "{}" ({}) is not.'.format(
                    key, list_of_text, type(list_of_text)
                )
            )
        if unique:
            list_of_text = [
                t for i, t in enumerate(list_of_text) if list_of_text.index(t) == i
            ]
    return list_of_text


def check_list_of_dicts(
    list_of_dicts: Optional[Iterable[Dict]],
    key: str,
    fields: Optional[List[str]] = None,
) -> Optional[List[Dict]]:
    """
    Validate iterable input to be a list/tuple/set of dicts.

    Optionally checking for required field names.

    :param list_of_dicts: list of dicts
    :param key: name of the object to display in the error message
    :param fields: list of fields that are required in each dict
    :raises IllegalArgumentError: if the list_of_dicts does not conform
    :returns: the list of dicts
    """
    if list_of_dicts is not None and list_of_dicts is not empty:
        if not isinstance(list_of_dicts, iter_types) or not all(
            isinstance(d, dict) for d in list_of_dicts
        ):
            raise IllegalArgumentError(
                '`{}` should be a list, tuple or set of dicts, "{}" ({}) is not.'.format(
                    key, list_of_dicts, type(list_of_dicts)
                )
            )
        if fields:
            assert isinstance(fields, list) and all(
                isinstance(f, str) for f in fields
            ), "`fields` must be a list of strings."
            missing_fields = set()
            for dictionary in list_of_dicts:
                for field in fields:
                    if field not in dictionary:
                        missing_fields.add(field)
            if missing_fields:
                raise IllegalArgumentError(
                    'Not every dict contains the required fields: "{}"\n'
                    'Missing fields: "{}"'.format(
                        '", "'.join(fields), '", "'.join(list(missing_fields))
                    )
                )
    return list_of_dicts


def check_enum(value: Optional[Any], enum: type(Enum), key: str) -> Optional[Any]:
    """Validate input to be an option from an enum class."""
    if value is not None and value is not empty:
        if value not in enum.values():
            raise IllegalArgumentError(
                '`{}` must be an option from enum {}, "{}" ({}) is not.\n'
                "Choose from: {}".format(
                    key, enum.__class__.__name__, value, type(value), enum.values()
                )
            )
    return value


def check_datetime(dt: Optional[Union[datetime, str]], key: str) -> Optional[str]:
    """Validate a datetime value to be a datetime and be timezone aware."""
    if dt is not None and dt is not empty:
        if isinstance(dt, str):
            dt = parse_datetime(dt)

        if isinstance(dt, datetime):
            if not dt.tzinfo:
                warnings.warn(
                    f"`{key}` `{dt.isoformat(sep=' ')}` is naive and not timezone aware, "
                    f"use pytz.timezone info. Date will be interpreted as UTC time."
                )
            dt = dt.isoformat(sep="T")
        else:
            raise IllegalArgumentError(
                f"`{key}` should be a correctly formatted string or a datetime.datetime() object, "
                f'"{dt}" ({type(dt)}) is not.'
            )
    return dt


def check_base(
    obj: Optional[Any],
    cls: Optional[type(object)] = None,
    key: Optional[str] = "object",
    method: Optional[Callable] = None,
) -> Optional[str]:
    """
    Validate whether the object provided is instance of Base (or subclass) and return its ID.

    When the obj is None, it returns None. Otherwise it will check if the object is a pykechain
    class and will extract the UUID from it.

    It will NOT check if the uuid (if a UUID is provided) is an actual class. It won't lookup
    the UUID in KE-chain and check that against the corresponding pykechain class.

    :param obj: Object that needs the checking.
    :param cls: (optional) See if the object is of a certain pykechain Class (subclass of Base)
    :param key: (optional) a key that may be provided to improve the legability of the provided
        Error
    :param method: (optional) a method or function that is used to convert the object into an
        instance that has the attribute 'id'.
    :returns: None or UUID
    :raises IllegalArgumentError: When the object is not of type of the class or not a UUID.
    """
    if obj is not None and obj is not empty:
        if cls is None:
            from pykechain.models import Base

            cls = Base

        if isinstance(obj, cls):
            obj = obj.id
        elif isinstance(obj, str) and is_uuid(obj):
            pass
        elif isinstance(obj, str) and method:
            obj = method(obj).id
        else:
            raise IllegalArgumentError(
                f'`{key}` must be an ID, UUID or `{cls.__name__}` object, "{obj}" ({type(obj)}) is'
                " not."
            )
    return obj


def check_user(
    obj: Optional[Any],
    cls: Optional[type(object)] = None,
    key: Optional[str] = "user",
    method: Optional[Callable] = None,
) -> Optional[int]:
    """Provide same functionality as check_base(), although users dont use UUID but integers."""
    if obj is not None and obj is not empty:

        if cls is None:
            from pykechain.models import User

            cls = User

        if isinstance(obj, cls):
            obj = obj.id
        elif isinstance(obj, (int, str)):
            try:
                obj = int(obj)
            except ValueError:
                raise IllegalArgumentError(
                    f'`{key}` must be an ID or `{cls.__name__}` object, "{obj}" ({type(obj)}) is'
                    " not."
                )
        elif method:
            obj = method(obj).id
        else:
            raise IllegalArgumentError(
                f'`{key}` must be an ID or `{cls.__name__}` object, "{obj}" ({type(obj)}) is not.'
            )
    return obj


def check_list_of_base(
    objects: Optional[List[Any]],
    cls: Optional[type(object)] = None,
    key: Optional[str] = "objects",
    method: Optional[Callable] = None,
) -> Optional[List[str]]:
    """
    Validate the iterable of objects are instance of Base (or subclass) and return a list of IDs.

    :param objects: list of objects to check
    :param cls: (optional) class to check objects against
    :param key: (optional) key to check
    :param method: (optional) method or function used to extract the `id` of the
        object (defaults `id`)
    :returns: list of UUID's in Text.
    """
    ids = None
    if objects is not None and objects is not empty:
        check_type(objects, iter_types, key=key)

        if cls is None:
            from pykechain.models import Base

            cls = Base

        ids = []
        not_recognized = []
        for obj in objects:
            try:
                ids.append(check_base(obj=obj, cls=cls, method=method))
            except IllegalArgumentError:
                not_recognized.append(obj)

        if not_recognized:
            raise IllegalArgumentError(
                'All `{}` must be IDs, UUIDs or `{}` objects, "{}" is/are not.'.format(
                    key, cls.__name__, '", "'.join([str(n) for n in not_recognized])
                )
            )
    return ids


def check_empty(value: Optional[Any]):
    """Validate whether the value provided is of class `Empty`."""
    if isinstance(value, Empty):
        return True
    return False



from typing import Dict, Optional

from pykechain.enums import (
    Alignment,
    SidebarAccessLevelOptions,
    SidebarItemAlignment,
    SidebarType,
    URITarget,
)
from pykechain.exceptions import IllegalArgumentError
from pykechain.models.sidebar.sidebar_base import SideBarItem


class SideBarCard(SideBarItem):
    """
    Side-bar card class.

    Every side-bar can have one or more cards and is maintained as an object of this class.

    :cvar allowed_attributes: allowed additional attributed provided as options alongside
        the specifically allowed ones.
    :cvar item_type: the item type of this class. Defaults to a BUTTON.
    """

    _allowed_attributes = [
        "displayText_nl",
        "displayText_en",
        "displayText_de",
        "displayText_fr",
        "displayText_it",
        "actionButtonName_nl",
        "actionButtonName_en",
        "actionButtonName_de",
        "actionButtonName_fr",
        "actionButtonName_it",
    ]

    _item_type = SidebarType.CARD

    def __init__(
        self,
        side_bar_manager: "SideBarManager",
        json: Optional[Dict] = None,
        alignment: SidebarItemAlignment = SidebarItemAlignment.TOP,
        minimum_access_level: SidebarAccessLevelOptions = SidebarAccessLevelOptions.IS_MEMBER,
        maximum_access_level: SidebarAccessLevelOptions = SidebarAccessLevelOptions.IS_MANAGER,
        display_text: str = None,
        show_close_action: bool = True,
        show_background: bool = True,
        show_action_button: bool = False,
        action_button_name: Optional[str] = None,
        action_button_uri: Optional[str] = None,
        action_button_uri_target: Optional[str] = None,
        display_text_align: Optional[Alignment] = Alignment.CENTER,
        **kwargs: dict,
    ) -> None:
        """
        Create a side-bar card.

        :param side_bar_manager: Manager object to which the button is linked.
        :param json: the json response to construct the :class:`SideBarButton` from
        :param title: visible label of the button
        :param icon: FontAwesome icon of the button
        :param uri: Uniform Resource Identifier, the address of the linked page
        :param uri_target: type of URI, either internal or external
        :param alignment: alignment of the button top or bottom
        :param minimum_access_level: the minimum permission needed to see the button
        :param maximum_access_level: the maximum permission needed to see the button
        :param icon_mode: FontAwesome display mode of the icon
        :returns None
        :raises IllegalArgumentError: When the provided Argument is not according to the type.
        """
        super().__init__()

        if json is None:
            json = {}
        self._manager: "SideBarManager" = side_bar_manager
        self.align = json.get("align", alignment)
        self.minimum_access_level = json.get("minimumAccessLevel", minimum_access_level)
        self.maximum_access_level = json.get("maximumAccessLevel", maximum_access_level)

        self.display_text = json.get("displayText", display_text)
        self.display_text_align = json.get("displayTextAlign", display_text_align)
        self.show_close_action = json.get("showCloseaction", show_close_action)
        self.show_background = json.get("showBackground", show_background)
        self.show_action_button = json.get("showActionButton", show_action_button)
        self.action_button_name = json.get("actionButtonName", action_button_name)
        self.action_button_uri = json.get("actionButtonUri", action_button_uri)
        self.action_button_uri_target = json.get(
            "actionButtonUriTarget", action_button_uri_target
        )

        if (
            self.action_button_uri_target is not None
            and self.action_button_uri_target not in URITarget.values()
        ):
            raise IllegalArgumentError(
                f'uri_target must be a URITarget option, "{self.action_button_uri_target}" is not.'
            )
        if self.align not in SidebarItemAlignment.values():
            raise IllegalArgumentError(
                f"alignment must be a proper `SidebarButtonAlgment` type, '{self.align} is not.'"
            )

        for key in kwargs.keys():
            if key not in self._allowed_attributes:
                raise IllegalArgumentError(
                    f'Attribute "{key}" is not supported in the configuration of a side-bar'
                    " card."
                )

        self._other_attributes = kwargs
        for key in self._allowed_attributes:
            if key in json:
                self._other_attributes[key] = json[key]

    def as_dict(self) -> Dict:
        """
        Retrieve the configuration data, or `meta`, of the side-bar button.

        :return: dictionary of the configuration data

        Example
        -------
        ```{
            "itemType": "CARD",
            "order": 5,
            "align": "top",
            "showCloseAction": true,
            "showActionButton": true,
            "actionButtonUri": "https://asdasdas",
            "actionButtonUriTarget": "_new",
            "actionButtonName": "Discover more",
            "displayText": "This project ...",
            "displayText_nl": "Dit project ...",
            "displayTextAlign": "left",
            "showBackground": true,
            "minimumAccessLevel": "is_member",
            "maximumAccessLevel": "is_supervisor",
        }```

        """
        config = {
            "itemType": self._item_type,
            "displayText": self.display_text,
            "displayTextAlign": self.display_text_align,
            "align": self.align,
            "minimumAccessLevel": self.minimum_access_level,
            "maximumAccessLevel": self.maximum_access_level,
            "showBackground": self.show_background,
            "showCloseAction": self.show_close_action,
            "showActionButton": self.show_action_button,
            "actionButtonName": self.action_button_name,
            "actionButtonUri": self.action_button_uri,
            "actionButtonUriTarget": self.action_button_uri_target,
        }
        config.update(self._other_attributes)
        config = {k: v for k, v in config.items() if v is not None}

        return config



import collections
import datetime
import os
from unittest import TestCase

import pytz
from betamax import Betamax

from pykechain import Client
from tests.utils import TEST_RECORD_CASSETTES, TEST_SCOPE_ID, TEST_SCOPE_NAME, TEST_TOKEN, TEST_URL

with Betamax.configure() as config:
    config.cassette_library_dir = os.path.join(os.path.dirname(__file__), "cassettes")
    config.define_cassette_placeholder("<API_URL>", TEST_URL)
    config.define_cassette_placeholder("<AUTH_TOKEN>", TEST_TOKEN)


class TestBetamax(TestCase):
    time = datetime.datetime(year=2020, month=1, day=1, tzinfo=pytz.UTC)

    @property
    def cassette_name(self):
        test = self._testMethodName
        return f"{self.__class__.__name__}.{test}"

    def setUp(self):
        # use self.env.set('var', 'value') and with self.env: ... to use custom environmental variables
        self.env = EnvironmentVarGuard()
        self.client = Client(url=TEST_URL)

        if TEST_TOKEN:
            self.client.login(token=TEST_TOKEN)

        self.recorder = Betamax(session=self.client.session)

        if TEST_RECORD_CASSETTES:
            self.recorder.use_cassette(self.cassette_name)
            self.recorder.start()
        else:
            self.recorder.config.record_mode = None

        if TEST_SCOPE_NAME:
            self.project = self.client.scope(name=TEST_SCOPE_NAME)
        elif TEST_SCOPE_ID:
            self.project = self.client.scope(id=TEST_SCOPE_ID)
        else:
            raise Exception(
                "Could not retrieve the test scope, you need to provide a "
                "`TEST_SCOPE_ID` or `TEST_SCOPE_NAME` in your `.env` file"
            )

    def tearDown(self):
        self.recorder.stop()
        del self.env
        del self.client


#
# This is EnvironmentVarGuard implementation of python 3.
# see: https://github.com/python/cpython/blob/3.10/Lib/test/support/os_helper.py#L562
#
class EnvironmentVarGuard(collections.abc.MutableMapping):
    """Class to help protect the environment variable properly.

    Can be used as a context manager."""

    def __init__(self):
        self._environ = os.environ
        self._changed = {}

    def __getitem__(self, envvar):
        return self._environ[envvar]

    def __setitem__(self, envvar, value):
        # Remember the initial value on the first access
        if envvar not in self._changed:
            self._changed[envvar] = self._environ.get(envvar)
        self._environ[envvar] = value

    def __delitem__(self, envvar):
        # Remember the initial value on the first access
        if envvar not in self._changed:
            self._changed[envvar] = self._environ.get(envvar)
        if envvar in self._environ:
            del self._environ[envvar]

    def keys(self):
        return self._environ.keys()

    def __iter__(self):
        return iter(self._environ)

    def __len__(self):
        return len(self._environ)

    def set(self, envvar, value):
        self[envvar] = value

    def unset(self, envvar):
        del self[envvar]

    def copy(self):
        # We do what os.environ.copy() does.
        return dict(self)

    def __enter__(self):
        return self

    def __exit__(self, *ignore_exc):
        for (k, v) in self._changed.items():
            if v is None:
                if k in self._environ:
                    del self._environ[k]
            else:
                self._environ[k] = v
        os.environ = self._environ



from typing import List, Optional, Union

import requests

from pykechain.defaults import API_EXTRA_PARAMS
from pykechain.enums import StatusCategory, TransitionType, WorkflowCategory
from pykechain.exceptions import APIError, IllegalArgumentError
from pykechain.models import Base, BaseInScope
from pykechain.models.base import CrudActionsMixin, NameDescriptionTranslationMixin
from pykechain.models.input_checks import (
    check_base,
    check_enum,
    check_list_of_base,
    check_text,
    check_type,
)
from pykechain.models.tags import TagsMixin
from pykechain.typing import ObjectID
from pykechain.utils import Empty, clean_empty_values, find_obj_in_list


class Transition(Base, CrudActionsMixin):
    """Transition Object."""

    url_detail_name = "transition"
    url_list_name = "transitions"
    url_pk_name = "transition_id"

    def __init__(self, json, **kwargs) -> None:
        """Initialise a Transition object."""
        super().__init__(json, **kwargs)
        self.description: str = json.get("description", "")
        self.ref: str = json.get("ref", "")
        self.derived_from_id: Optional[ObjectID] = json.get("derived_from")

        self.from_status: List[Status] = [
            Status(j, client=self._client) for j in json.get("grom_status", [])
        ]
        self.to_status: Status = json.get("to_status")
        self.transition_type: TransitionType = json.get("transition_type")

        self.conditions: dict = json.get("conditions", {})
        self.validators: dict = json.get("validators", {})
        self.post_functions: dict = json.get("post_functions", {})
        self.transition_screen_id: Optional[ObjectID] = json.get("transition_screen")

    @classmethod
    def list(cls, client: "Client", **kwargs) -> List["Transition"]:
        """Retrieve a list of Transition objects through the client."""
        return super().list(client=client, **kwargs)

    @classmethod
    def get(cls, client: "Client", **kwargs) -> "Transition":
        """Retrieve a Transition object through the client."""
        return super().get(client=client, **kwargs)


class Status(Base, CrudActionsMixin):
    """Status object."""

    url_detail_name = "status"
    url_list_name = "statuses"
    url_pk_name = "status_id"

    def __init__(self, json, **kwargs) -> None:
        """Initialize a Status object."""
        super().__init__(json, **kwargs)
        self.description: str = json.get("description", "")
        self.ref: str = json.get("ref", "")
        self.status_category: StatusCategory = json.get("status_category")

    @classmethod
    def list(cls, client: "Client", **kwargs) -> List["Status"]:
        """Retrieve a list of Status objects through the client."""
        return super().list(client=client, **kwargs)

    @classmethod
    def get(cls, client: "Client", **kwargs) -> "Status":
        """Retrieve a Status object through the client."""
        return super().get(client=client, **kwargs)


class Workflow(
    BaseInScope, CrudActionsMixin, TagsMixin, NameDescriptionTranslationMixin
):
    """Workflow object."""

    url_detail_name = "workflow"
    url_list_name = "workflows"
    url_pk_name: str = "workflow_id"

    def __init__(self, json, **kwargs) -> None:
        """Initialize a Workflow Object."""
        super().__init__(json, **kwargs)
        self.description: str = json.get("description", "")
        self.ref: str = json.get("ref", "")
        self.derived_from_id: Optional[ObjectID] = json.get("derived_from_id")
        self._transitions: List[Transition] = [
            Transition(j, client=self._client) for j in json.get("transitions", [])
        ]
        self.category: WorkflowCategory = json.get("category")
        self.options: dict = json.get("options", {})
        self.active: bool = json.get("active")
        self._statuses: List[Status] = [
            Status(j, client=self._client) for j in json.get("statuses")
        ]

    def __repr__(self) -> str:  # pragma: no cover
        return f"<pyke Workflow '{self.name}' '{self.category}' id {self.id[-8:]}>"

    def edit(
        self, name: str = Empty(), description: str = Empty(), *args, **kwargs
    ) -> None:
        """Change the workflow object.

        Change the name and description of a workflow. It is also possible to update the workflow
        options and also the 'active' flag. To change the active flag of the workflow we
        kindly refer to the `activate()` and `deactivate()` methods on the workflow.

        :type name: name of the workflow is required.
        :type description: (optional) description of the workflow

        """
        if isinstance(name, Empty):
            name = self.name
        data = {
            "name": check_text(name, "name"),
            "description": check_text(description, "description"),
            "active": check_type(kwargs.get("active", Empty()), bool, "active"),
            "options": check_type(kwargs.get("options", Empty()), bool, "options"),
        }
        url = self._client._build_url("workflow", workflow_id=self.id)
        query_params = API_EXTRA_PARAMS.get(self.url_list_name)
        response = self._client._request(
            "PUT",
            url=url,
            params=query_params,
            json=clean_empty_values(data, nones=False),
        )
        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError("Could not clone the workflow", response=response)
        self.refresh(json=response.json()["results"][0])

    @classmethod
    def list(cls, client: "Client", **kwargs) -> List["Workflow"]:
        """Retrieve a list of Workflow objects through the client."""
        return super().list(client=client, **kwargs)

    @classmethod
    def get(cls, client: "Client", **kwargs) -> "Workflow":
        """Retrieve a single Workflow object using the client."""
        return super().get(client=client, **kwargs)

    def delete(self):
        """Delete Workflow."""
        return super().delete()

    @property
    def status_order(self) -> List[Status]:
        """Statuses in the right order."""
        return self.statuses

    @status_order.setter
    def status_order(self, value: List[Union[ObjectID, Status]]):
        """
        Set the order of the statuses specifically in a certain order.

        :param value: a determined ordered list of Status or status UUID's
        """
        if not value:
            raise IllegalArgumentError(
                f"To set the order of statises, provide a list of status objects, got: '{value}'"
            )
        data = {"status_order": check_list_of_base(value, Status, "statuses")}
        url = self._client._build_url("workflow_set_status_order", workflow_id=self.id)
        response = self._client._request("PUT", url=url, json=data)
        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError(
                "Could not alter the order of the statuses", response=response
            )
        self.refresh(json=response.json()["results"][0])

    #
    # Subclass finders and managers.
    #

    def transition(
        self,
        value: str = None,
        attr: str = None,
    ) -> Transition:
        """
        Retrieve the Transition belonging to this workflow based on its name, ref or uuid.

        :param value: transition name, ref or UUID to search for
        :param attr: the attribute to match on. Eg. to_status=<Status Obj>
        :return: a single :class:`Transition`
        :raises NotFoundError: if the `Transition` is not part of the `Workflow`
        :raises MultipleFoundError

        Example
        -------
        >>> workflow = project.workflow('Simple Flow')
        >>> transition = workflow.transition('in progress')
        >>> todo_status = client.status(name="To Do")
        >>> transition = workflow.transition(todo_status, attr="to_status")

        """
        return find_obj_in_list(value, iterable=self._transitions, attribute=attr)

    @property
    def transitions(self):
        """
        Retrieve the Transitions belonging to this workflow.

        :return: multiple :class:`Transition`
        """
        return self._transitions

    def status(self, value: str = None, attr: str = None) -> Status:
        """
        Retrieve the Status belonging to this workflow based on its name, ref or uuid.

        :param value: status name, ref or UUID to search for
        :param attr: the attribute to match on.
        :return: a single :class:`Status`
        :raises NotFoundError: if the `Status` is not part of the `Workflow`
        :raises MultipleFoundError

        Example
        -------
        >>> workflow = project.workflow('Simple Flow')
        >>> status = workflow.status('To Do')

        """
        return find_obj_in_list(value, iterable=self._statuses, attribute=attr)

    @property
    def statuses(self):
        """
        Retrieve the Statuses belonging to this workflow.

        :return: multiple :class:`Status`
        """
        return self._statuses

    #
    # Mutable methods on the object
    #

    def activate(self):
        """Set the active status to True."""
        if not self.active:
            url = self._client._build_url("workflow_activate", workflow_id=self.id)
            response = self._client._request("PUT", url=url)
            if response.status_code != requests.codes.ok:  # pragma: no cover
                raise APIError("Could not activate the workflow", response=response)

            # we need to do a full refresh here from the server as the
            # API of workflow/<id>/activate does not return the full object as response.
            self.refresh(
                url=self._client._build_url("workflow", workflow_id=self.id),
                extra_params=API_EXTRA_PARAMS.get(self.url_list_name),
            )

    def deactivate(self):
        """Set the active status to False."""
        if self.active:
            url = self._client._build_url("workflow_deactivate", workflow_id=self.id)
            response = self._client._request("PUT", url=url)
            if response.status_code != requests.codes.ok:  # pragma: no cover
                raise APIError("Could not activate the workflow", response=response)

            # we need to do a full refresh here from the server as the
            # API of workflow/<id>/deactivate does not return the full object as response.
            self.refresh(
                url=self._client._build_url("workflow", workflow_id=self.id),
                extra_params=API_EXTRA_PARAMS.get(self.url_list_name),
            )

    def clone(
        self,
        target_scope: "Scope" = Empty(),
        name: Optional[str] = Empty(),
        description: Optional[str] = Empty(),
    ) -> "Workflow":
        """Clone the current workflow into a new workflow.

        Also used to 'import' a catalog workflow into a scope.

        :param target_scope: (optional) target scope where to clone the Workflow to.
            Defaults current scope.
        :param name: (optional) name of the new workflow
        :param description: (optional) description of the new workflow
        """
        from pykechain.models import Scope

        if isinstance(target_scope, Empty):
            target_scope = self.scope_id
        data = {
            "target_scope": check_base(target_scope, Scope, "scope"),
            "name": check_text(name, "name"),
            "description": check_text(description, "description"),
        }
        url = self._client._build_url("workflow_clone", workflow_id=self.id)
        query_params = API_EXTRA_PARAMS.get(self.url_list_name)
        response = self._client._request(
            "POST",
            url=url,
            params=query_params,
            json=clean_empty_values(data, nones=False),
        )
        if response.status_code != requests.codes.created:  # pragma: no cover
            raise APIError("Could not clone the workflow", response=response)
        return Workflow(json=response.json()["results"][0], client=self._client)

    def update_transition(
        self,
        transition: Union[Transition, ObjectID],
        name: Optional[str] = Empty(),
        description: Optional[str] = Empty(),
        from_status: Optional[List[str]] = Empty(),
    ) -> Transition:
        """Update a specific Transition in the current workflow.

        Update the transition inside the worfklow based on a transition_id.

        :param transition: Transition object or transition id to alter
        :param name: (optional) name to change
        :param description: (optional) description to change
        :param from_status: (optional) a list of from statuses to update
        """
        data = {
            "transition_id": check_base(transition, Transition, "transition_id"),
            "name": check_text(name, "name"),
            "decription": check_text(description, "description"),
            "from_status": check_list_of_base(from_status, Status, "from_statuses"),
        }
        url = self._client._build_url("workflow_update_transition", workflow_id=self.id)
        query_params = API_EXTRA_PARAMS.get(self.url_list_name)
        response = self._client._request(
            "POST", url=url, params=query_params, json=data
        )
        if response.status_code != requests.codes.ok:
            raise APIError(
                f"Could not update the specific transition '{transition}' in the "
                "workflow",
                response=response,
            )
        # an updated transition will be altered so we want to refresh the workflow.
        self.refresh()
        return Transition(json=response.json()["results"][0])

    def delete_transition(self, transition: Union[Transition, ObjectID]) -> None:
        """Remove Transition from the current Workflow and delete it.

        If the Transition is still connected to *other* Workflows, it will *not* be
        removed, and will result in a 400 reporting all attached Workflows.

        :param transition: object or uuid of a transition to delete.
        """
        transition_id = check_base(transition, Transition, "transition_id")
        url = self._client._build_url(
            "workflow_delete_transition",
            workflow_id=self.id,
            transition_id=transition_id,
        )
        query_params = API_EXTRA_PARAMS.get(self.url_list_name)
        response = self._client._request("DELETE", url=url, params=query_params)
        if response.status_code != requests.codes.no_content:
            raise APIError(
                f"Could not delete the specific transition '{transition}' from the "
                "workflow",
                response=response,
            )
        # a deleted transition will be unlinked so we want to refresh the workflow.
        self.refresh()

    def create_transition(
        self,
        name: str,
        to_status: Union[Status, ObjectID],
        transition_type: TransitionType = TransitionType.GLOBAL,
        from_status: Optional[List[Union[Status, ObjectID]]] = Empty(),
        description: Optional[str] = Empty(),
    ):
        """
        Create a new Transition and associate it to the current Workflow.

        :param name: name of the transition.
        :param to_status: status where to transition to (a single status or status id)
        :param transition_type: transition type to transition to. (defaults to Global)
        :param from_status: (optional) status to transition from. Not used for Global transitions.
        :param description: (optional) description.
        """
        data = {
            "name": check_text(name, "name"),
            "to_status": check_base(to_status, Status, "to_status"),
            "transition_type": check_enum(
                transition_type, TransitionType, "transition_type"
            ),
            "from_status": check_list_of_base(from_status, Status, "from_statuses"),
            "description": check_text(description, "description"),
        }
        url = self._client._build_url("workflow_create_transition", workflow_id=self.id)
        query_params = API_EXTRA_PARAMS.get(self.url_list_name)
        response = self._client._request(
            "POST", url=url, params=query_params, json=clean_empty_values(data)
        )
        if response.status_code != requests.codes.created:
            raise APIError(
                "Could not create the specific transition in the " "workflow",
                response=response,
            )
        # a new transition will be linked to the workflow so we want to refresh the workflow.
        self.refresh()
        return Transition(json=response.json()["results"][0], client=self._client)

    def create_status(
        self,
        name: str,
        category: StatusCategory = StatusCategory.UNDEFINED,
        description: Optional[str] = Empty(),
    ) -> Status:
        """Create a new Status.

        Will create a new status, a new global transition to that status and
        will link the new Global transition to that status to the current
        workflow.

        :param name: name of the status
        :param category: status category (defaults to UNDEFINED)
        :param description: (optional) description of the status
        """
        data = {
            "name": check_text(name, "name"),
            "status_category": check_enum(category, StatusCategory, "status_category"),
            "description": check_text(description, "description"),
        }

        url = self._client._build_url("workflow_create_status", workflow_id=self.id)
        query_params = API_EXTRA_PARAMS.get(self.url_list_name)
        response = self._client._request(
            "POST", url=url, params=query_params, json=clean_empty_values(data)
        )
        if response.status_code != requests.codes.created:
            raise APIError(
                "Could not create the specific status, a global transition and"
                "link it to the workflow",
                response=response,
            )
        # a new status will create a new global transition to that status
        # so we want to update the current workflow.
        self.refresh()
        return Status(json=response.json()["results"][0], client=self._client)

    def link_transitions(self, transitions: List[Union[Transition, ObjectID]]):
        """
        Link a list of Transitions to a Workflow.

        :param transitions: a list of Transition Objects or transition_ids to link to the workflow.
        """
        data = {"transitions": check_list_of_base(transitions, Transition)}
        url = self._client._build_url("workflow_link_transitions", workflow_id=self.id)
        query_params = API_EXTRA_PARAMS.get(self.url_list_name)
        response = self._client._request(
            "POST", url=url, params=query_params, json=data
        )
        if response.status_code != requests.codes.ok:
            raise APIError(
                "Could not create the specific status, a global transition and"
                "link it to the workflow",
                response=response,
            )
        self.refresh(json=response.json()["results"][0])

    def unlink_transitions(
        self, transitions: List[Union[Transition, ObjectID]]
    ) -> None:
        """
        Unlink a list of Transitions to a Workflow.

        :param transitions: a list of Transition Objects or transition_ids to link to the workflow.
        """
        data = {"transitions": check_list_of_base(transitions, Transition)}
        url = self._client._build_url(
            "workflow_unlink_transitions", workflow_id=self.id
        )
        query_params = API_EXTRA_PARAMS.get(self.url_list_name)
        response = self._client._request(
            "POST", url=url, params=query_params, json=data
        )
        if response.status_code != requests.codes.ok:
            raise APIError(
                "Could not create the specific status, a global transition and"
                "link it to the workflow",
                response=response,
            )
        self.refresh(json=response.json()["results"][0])



from copy import deepcopy
from random import shuffle
from unittest import skip

from pykechain.enums import StatusCategory, TransitionType, WorkflowCategory
from pykechain.exceptions import ForbiddenError, NotFoundError
from pykechain.models.input_checks import check_list_of_base
from pykechain.models.workflow import Status, Workflow
from tests.classes import TestBetamax


class TestWorkflows(TestBetamax):
    """
    Test retrieval and scope attributes and methods.
    """

    def setUp(self):
        super().setUp()
        self.workflow: Workflow = self.client.workflow(
            "Simple Workflow", category=WorkflowCategory.CATALOG
        )

    def tearDown(self):
        super().tearDown()

    def test_workflow_attributes(self):
        attributes = [
            "id",
            "name",
            "description",
            "ref",
            "derived_from_id",
            "created_at",
            "updated_at",
            "category",
            "scope_id",
            "statuses",
            "transitions",
            "options",
            "active",
        ]

        workflow = self.client.workflow(
            "Simple Workflow", category=WorkflowCategory.CATALOG
        )
        for attr in attributes:
            with self.subTest(attr):
                self.assertTrue(
                    hasattr(workflow, attr),
                    f"Could not find '{attr}' in the workflow: '{workflow.__dict__.keys()}",
                )

    def test_single_workflow_retrieve_on_pk(self):
        workflow_id = self.workflow.id

        retr_on_id = self.client.workflow(pk=workflow_id)
        pass

    def test_clone_workflow_to_scope(self):
        """Testing the 'import workflow'."""
        catalog_workflow = self.client.workflow(
            name="Simple Workflow", category=WorkflowCategory.CATALOG
        )

        imported_workflow = catalog_workflow.clone(
            target_scope=self.project, name="___Imported Simple Workflow"
        )

        self.assertTrue(imported_workflow)
        retrieved_imported_workflow = self.client.workflow(
            name="___Imported Simple Workflow", scope=self.project
        )
        self.assertEqual(imported_workflow.id, retrieved_imported_workflow.id)

        retrieved_imported_workflow.delete()

    def test_import_workflow_in_a_scope_with_import_workflow_method_on_scope(self):
        """Test the create of a new workflow in a scope."""
        imported_workflow = self.project.import_workflow(
            workflow=self.workflow, name="___Imported Simple Flow"
        )
        self.assertEqual(imported_workflow.derived_from_id, self.workflow.id)
        imported_workflow.delete()

    def test_clone_workflow_in_a_scope_in_the_same_scope(self):
        test_wf = self.project.import_workflow(workflow=self.workflow, name="__TEST")

        cloned_wf = test_wf.clone(
            target_scope=test_wf.scope_id, name="CLONED WITHIN THE SOCPE"
        )
        self.assertTrue(cloned_wf)
        self.assertEqual(cloned_wf.derived_from_id, test_wf.id)

        # teardown
        test_wf.delete()
        cloned_wf.delete()


class TestWorkflowMethods(TestBetamax):
    def setUp(self):
        super().setUp()
        self.catalog_workflow: Workflow = self.client.workflow(
            "Simple Workflow", category=WorkflowCategory.CATALOG
        )
        self.workflow = self.project.import_workflow(
            workflow=self.catalog_workflow,
            name="__TEST Imported Simple Flow from Catalog",
        )

    def tearDown(self):
        super().tearDown()
        try:
            self.workflow.delete()
        except (NotFoundError, ForbiddenError):
            # we pass this silently as workflow delete when the
            # cassettes are irrelevant and returns a 404 NotFoundError
            pass

    def test_edit_workflow_description(self):
        new_description = "foo bar"
        self.workflow.edit(
            description=new_description
        )
        self.assertEqual(self.workflow.description, new_description)


    def test_workflow_in_and_activate(self):
        self.workflow.deactivate()
        self.assertFalse(self.workflow.active)
        self.workflow.activate()
        self.assertTrue(self.workflow.active)

    def test_create_delete_status_and_create_delete_workflow_transition(self):
        """Test to create a transition on a workflow.

        - tests creation of status on workflow
        - tests deletion of status
        - tests deletion of transition on workflow
        """

        count_transitions = len(self.workflow.transitions)
        test_status = self.workflow.create_status(
            name="__TEST DONE", category=StatusCategory.DONE
        )
        test_transition = self.workflow.transition(test_status.name)
        self.workflow.delete_transition(test_transition)
        from_status = self.workflow.create_status(
            name="__TEST FROM", category=StatusCategory.TODO
        )
        from_transition = self.workflow.transition(from_status.name)
        self.workflow.delete_transition(from_transition)

        new_transition = self.workflow.create_transition(
            name="___test transition",
            to_status=test_status,
            transition_type=TransitionType.DIRECTED,
            from_status=[from_status],
        )

        self.workflow.delete_transition(new_transition)
        test_status.delete()
        from_status.delete()

    def test_workflow_status_order(self):
        cloned_wf = self.workflow.clone(
            name="___cloned_workflow_status_order", target_scope=self.workflow.scope_id
        )

        try:
            status_ids = [s.id for s in deepcopy(cloned_wf.statuses)]
            reversed_status_ids = list(reversed(status_ids))

            cloned_wf.status_order = reversed_status_ids

            self.assertListEqual(
                reversed_status_ids, check_list_of_base(cloned_wf.status_order, Status)
            )
        finally:
            cloned_wf.delete()

    def test_workflow_unlink_and_link_transitions(self):
        transitions = deepcopy(self.workflow.transitions)
        the_transition_to_unlink = transitions[-1]

        # unlink
        self.workflow.unlink_transitions([the_transition_to_unlink])
        self.assertListEqual(self.workflow.transitions, transitions[:-1])

        # relink
        self.workflow.link_transitions([the_transition_to_unlink])
        self.assertListEqual(self.workflow.transitions, transitions)



"""Side-bar buttons manager.

.. versionadded:: 3.x.x
"""
from pykechain.models.sidebar.sidebar_button import SideBarButton
from pykechain.models.sidebar.sidebar_card import SideBarCard
from pykechain.models.sidebar.sidebar_manager import SideBarManager

__all__ = (
    "SideBarManager",
    "SideBarButton",
    "SideBarCard"
)



from pykechain.utils import __dict__inherited__


class Enum:
    """Custom enumeration class to support class attributes as options.

    Example
    -------
    >>> class Toppings(Enum):
    ...    CHEESE = "Cheese"
    ...    SALAMI = "Salami"
    >>> topping_choice = Toppings.CHEESE

    """

    @classmethod
    def options(cls):
        """Provide a sorted list of options."""
        return sorted(
            (value, name)
            for (name, value) in __dict__inherited__(cls=cls, stop=Enum).items()
        )

    @classmethod
    def values(cls):
        """Provide a (sorted) list of values."""
        return [value for (value, name) in cls.options()]


class Multiplicity(Enum):
    """The various multiplicities that are accepted by KE-chain.

    For more information on the representation in KE-chain, please consult the KE-chain
    `Part documentation`_.

    :cvar ZERO_ONE: Multiplicity 0 to 1
    :cvar ONE: Multiplicity 1
    :cvar ZERO_MANY: Multiplicity 0 to infinity
    :cvar ONE_MANY: Multiplicity 1 to infinity
    """

    ZERO_ONE = "ZERO_ONE"
    ONE = "ONE"
    ZERO_MANY = "ZERO_MANY"
    ONE_MANY = "ONE_MANY"
    # M_N = "M_N"  # not implemented


class Category(Enum):
    """The various categories of Parts that are accepted by KE-chain.

    For more information on the representation in KE-chain, please consult the KE-chain
    `Part documentation`_.

    :cvar INSTANCE: Category of Instance
    :cvar MODEL: Category of Model
    """

    INSTANCE = "INSTANCE"
    MODEL = "MODEL"


class Classification(Enum):
    """The various classification of Parts that are accepted by KE-chain.

    For more information on the representation in KE-chain, please consult the KE-chain
    `Part documentation`_.

    :cvar PRODUCT: Classification of the part object is Product
    :cvar CATALOG: Classification of the part object is a CATALOG

    .. _Part documentation: https://support.ke-chain.com/confluence/dosearchsite.action?
            queryString=concept+part
    """

    PRODUCT = "PRODUCT"
    CATALOG = "CATALOG"
    FORM = "FORM"


class PropertyType(Enum):
    """The various property types that are accepted by KE-chain.

    For more information on the representation in KE-chain, please consult the KE-chain
    `Property documentation`_.

    :cvar CHAR_VALUE: a charfield property (single line text)
    :cvar TEXT_VALUE: text property (long text, may span multiple lines)
    :cvar BOOLEAN_VALUE: a boolean value property (True/False)
    :cvar INT_VALUE: integer property (whole number)
    :cvar FLOAT_VALUE: floating point number property (with digits)
    :cvar DATETIME_VALUE: a datetime value property
    :cvar ATTACHMENT_VALUE: an attachment property
    :cvar LINK_VALUE: url property
    :cvar REFERENCE_VALUE: a reference property, a UUID value referring to other part model

    .. versionadded:: 1.14

    :cvar SINGLE_SELECT_VALUE: single select list property (choose from a list)
    :cvar REFERENCES_VALUE: a multi reference property, a list of UUID values referring to
        other part models

    .. versionadded:: 3.6
    :cvar ACTIVITY_REFERENCES_VALUE: Activity References Property
    :cvar SCOPE_REFERENCES_VALUE: Scope References Property
    :cvar SERVICE_REFERENCES_VALUE: Service Referenes Property
    :cvar TEAM_REFERENCES_VALUE: Team References Property
    :cvar USER_REFERENCES_VALUE: User References Property
    :cvar FORM_REFERENCES_VALUE: Form References Property
    :cvar CONTEXT_REFERENCES_VALUE: Context References Property
    :cvar JSON_VALUE: Generic JSON storage Property
    :cvar GEOJSON_VALUE: GEOJSON property to store map data
    :cvar WEATHER_VALUE: Weather JSON property compatible with the response of weatherapi.com

    .. versionadded:: 3.19
    :cvar STATUS_REFERENCES_VALUE: Status References Property

    .. _Property documentation: https://support.ke-chain.com/confluence/dosearchsite.action?
        queryString=concept+property
    """

    CHAR_VALUE = "CHAR_VALUE"
    TEXT_VALUE = "TEXT_VALUE"
    BOOLEAN_VALUE = "BOOLEAN_VALUE"
    INT_VALUE = "INT_VALUE"
    FLOAT_VALUE = "FLOAT_VALUE"
    DATETIME_VALUE = "DATETIME_VALUE"
    DATE_VALUE = "DATE_VALUE"
    TIME_VALUE = "TIME_VALUE"
    ATTACHMENT_VALUE = "ATTACHMENT_VALUE"
    LINK_VALUE = "LINK_VALUE"
    SINGLE_SELECT_VALUE = "SINGLE_SELECT_VALUE"
    MULTI_SELECT_VALUE = "MULTI_SELECT_VALUE"
    REFERENCE_VALUE = "REFERENCE_VALUE"
    REFERENCES_VALUE = "REFERENCES_VALUE"
    ACTIVITY_REFERENCES_VALUE = "ACTIVITY_REFERENCES_VALUE"
    SCOPE_REFERENCES_VALUE = "SCOPE_REFERENCES_VALUE"
    SERVICE_REFERENCES_VALUE = "SERVICE_REFERENCES_VALUE"
    FORM_REFERENCES_VALUE = "FORM_REFERENCES_VALUE"
    CONTEXT_REFERENCES_VALUE = "CONTEXT_REFERENCES_VALUE"
    STATUS_REFERENCES_VALUE = "STATUS_REFERENCES_VALUE"
    TEAM_REFERENCES_VALUE = "TEAM_REFERENCES_VALUE"
    USER_REFERENCES_VALUE = "USER_REFERENCES_VALUE"
    JSON_VALUE = "JSON_VALUE"
    GEOJSON_VALUE = "GEOJSON_VALUE"
    WEATHER_VALUE = "WEATHER_VALUE"


class ActivityType(Enum):
    """The various Activity types that are accepted by KE-chain.

    .. versionadded:: 2.0

    :cvar TASK: a normal task
    :cvar PROCESS: a subprocess (container) containing other tasks
    """

    PROCESS = "PROCESS"
    TASK = "TASK"


class ActivityClassification(Enum):
    """The classification of Activities that are accepted by KE-chain.

    .. versionadded:: 2.0
    .. versionchanged:: 3.2
       Add 'APP' environment for KE-chain versions > 3.1
    .. versionchanged:: 3.14
       Add 'FORM' envornment for KE-chain versions > v2021.10


    :cvar WORKFLOW: Classification of the activity is WORKFLOW
    :cvar CATALOG: Classification of the activity is CATALOG
    :cvar APP: Classification of the activity is APP
    :cvar FORM: Classification of the activity is FORM
    """

    WORKFLOW = "WORKFLOW"
    CATALOG = "CATALOG"
    APP = "APP"
    FORM = "FORM"


class ActivityRootNames(Enum):
    """The classification of Activities that are accepted by KE-chain.

    .. versionadded:: 2.0
    .. versionchanged:: 3.2
       Add 'APP' environment for KE-chain versions > 3.1
    .. versionchanged:: 3.14
       Add 'FORM' environment for KE-chain versions >= v2021.10

    :cvar WORKFLOW_ROOT: Root of the activity is WORKFLOW_ROOT
    :cvar CATALOG_ROOT: Root of the activity is CATALOG_ROOT (below are CATALOG tasks)
    :cvar APP_ROOT: Root of the activity is APP_ROOT (below are APP 'tasks' ie. 'screems')
    """

    WORKFLOW_ROOT = "WORKFLOW_ROOT"
    CATALOG_ROOT = "CATALOG_ROOT"
    APP_ROOT = "APP_ROOT"
    FORM_ROOT = "FORM_ROOT"


activity_root_name_by_classification = {
    ActivityClassification.WORKFLOW: ActivityRootNames.WORKFLOW_ROOT,
    ActivityClassification.CATALOG: ActivityRootNames.CATALOG_ROOT,
    ActivityClassification.APP: ActivityRootNames.APP_ROOT,
    ActivityClassification.FORM: ActivityRootNames.FORM_ROOT,
}


class WidgetNames(Enum):
    """The various Names of the Widget that can be configured.

    .. versionchanged:: 3.14
       Added FORMMETAPANEL for KE-chain versions >= v2021.10

    :cvar SUPERGRIDWIDGET: superGridWidget
    :cvar PROPERTYGRIDWIDGET: propertyGridWidget
    :cvar HTMLWIDGET: htmlWidget
    :cvar FILTEREDGRIDWIDGET: filteredGridWidget
    :cvar SERVICEWIDGET: serviceWidget
    :cvar NOTEBOOKWIDGET: notebookWidget
    :cvar ATTACHMENTVIEWERWIDGET: attachmentViewerWidget
    :cvar TASKNAVIGATIONBARWIDGET: taskNavigationBarWidget
    :cvar JSONWIDGET: jsonWidget

    # KE-chain 3 only
    :cvar SIGNATUREWIDGET: signatureWidget
    :cvar CARDWIDGET: cardWidget
    :cvar METAPANELWIDGET: metaPanelWidget
    :cvar FORMMETAPANEL: formMetaPanelWidget
    :cvar MULTICOLUMNWIDGET: multiColumnWidget
    :cvar PROGRESSWIDGET: progressWidget
    :cvar TASKSWIDGET: tasksWidget
    :cvar SERVICECARDWIDGET: serviceCardWidget
    :cvar DASHBOARDWIDGET: 'dashboardWidget'
    :cvar SCOPEMEMBERS: 'scopeMembersWidget'
    """

    SUPERGRIDWIDGET = "superGridWidget"
    PROPERTYGRIDWIDGET = "propertyGridWidget"
    HTMLWIDGET = "htmlWidget"
    FILTEREDGRIDWIDGET = "filteredGridWidget"
    SERVICEWIDGET = "serviceWidget"
    NOTEBOOKWIDGET = "notebookWidget"
    ATTACHMENTVIEWERWIDGET = "attachmentViewerWidget"
    TASKNAVIGATIONBARWIDGET = "taskNavigationBarWidget"
    JSONWIDGET = "jsonWidget"
    METAPANELWIDGET = "metaPanelWidget"
    FORMMETAPANELWIDGET = "formMetaPanelWidget"
    MULTICOLUMNWIDGET = "multiColumnWidget"
    SIGNATUREWIDGET = "signatureWidget"
    CARDWIDGET = "cardWidget"
    PROGRESSWIDGET = "progressWidget"
    TASKSWIDGET = "taskWidget"
    SERVICECARDWIDGET = "serviceCardWidget"
    DASHBOARDWIDGET = "dashboardWidget"
    SCOPEMEMBERS = "scopeMembersWidget"


class WidgetTypes(Enum):
    """The various widget types for the widget definitions available to the widget api.

    .. versionchanged:: 3.14
       Added FORMMETAPANEL for KE-chain versions >= v2021.10

    :cvar UNDEFINED: Undefined Widget
    :cvar PROPERTYGRID: Propertygrid widget
    :cvar SUPERGRID: Supergrid widget
    :cvar HTML: Html widget
    :cvar FILTEREDGRID: Filteredgrid widget
    :cvar SERVICE: Service widget
    :cvar NOTEBOOK: Notebook widget
    :cvar ATTACHMENTVIEWER: Attachmentviewer widget
    :cvar TASKNAVIGATIONBAR: Tasknavigationbar widget
    :cvar JSON: Json widget
    :cvar METAPANEL: Metapanel widget
    :cvar FORMMETAPANEL: The FormMetapanel widget
    :cvar MULTICOLUMN: Multicolumn widget
    :cvar SCOPE: Scope widget
    :cvar THIRDPARTY: Thirdparty widget
    :cvar PROGRESS: Progress widget
    :cvar SIGNATURE: Signature widget
    :cvar CARD: Card widget
    :cvar TASKS: Tasks widget
    :cvar WEATHER: Weather widget
    :cvar SERVICECARD: Servicecard widget
    :cvar DASHBOARD: Dashboard widget
    :cvar SCOPEMEMBERS: Scopemembers widget
    """

    UNDEFINED = "UNDEFINED"
    PROPERTYGRID = "PROPERTYGRID"
    SUPERGRID = "SUPERGRID"
    HTML = "HTML"
    FILTEREDGRID = "FILTEREDGRID"
    SERVICE = "SERVICE"
    NOTEBOOK = "NOTEBOOK"
    ATTACHMENTVIEWER = "ATTACHMENTVIEWER"
    TASKNAVIGATIONBAR = "TASKNAVIGATIONBAR"
    JSON = "JSON"
    METAPANEL = "METAPANEL"
    FORMMETAPANEL = "FORMMETAPANEL"
    MULTICOLUMN = "MULTICOLUMN"
    SCOPE = "SCOPE"
    THIRDPARTY = "THIRDPARTY"
    PROGRESS = "PROGRESS"
    SIGNATURE = "SIGNATURE"
    CARD = "CARD"
    TASKS = "TASKS"
    WEATHER = "WEATHER"
    SERVICECARD = "SERVICECARD"
    DASHBOARD = "DASHBOARD"
    SCOPEMEMBERS = "SCOPEMEMBERS"


WidgetCompatibleTypes = {
    WidgetNames.SUPERGRIDWIDGET: WidgetTypes.SUPERGRID,
    WidgetNames.PROPERTYGRIDWIDGET: WidgetTypes.PROPERTYGRID,
    WidgetNames.HTMLWIDGET: WidgetTypes.HTML,
    WidgetNames.FILTEREDGRIDWIDGET: WidgetTypes.FILTEREDGRID,
    WidgetNames.SERVICEWIDGET: WidgetTypes.SERVICE,
    WidgetNames.NOTEBOOKWIDGET: WidgetTypes.NOTEBOOK,
    WidgetNames.ATTACHMENTVIEWERWIDGET: WidgetTypes.ATTACHMENTVIEWER,
    WidgetNames.TASKNAVIGATIONBARWIDGET: WidgetTypes.TASKNAVIGATIONBAR,
    WidgetNames.JSONWIDGET: WidgetTypes.JSON,
    WidgetNames.METAPANELWIDGET: WidgetTypes.METAPANEL,
    WidgetNames.FORMMETAPANELWIDGET: WidgetTypes.FORMMETAPANEL,
    WidgetNames.MULTICOLUMNWIDGET: WidgetTypes.MULTICOLUMN,
    WidgetNames.PROGRESSWIDGET: WidgetTypes.PROGRESS,
    WidgetNames.SIGNATUREWIDGET: WidgetTypes.SIGNATURE,
    WidgetNames.CARDWIDGET: WidgetTypes.CARD,
    WidgetNames.TASKSWIDGET: WidgetTypes.TASKS,
    WidgetNames.SERVICECARDWIDGET: WidgetTypes.SERVICECARD,
    WidgetNames.DASHBOARDWIDGET: WidgetTypes.DASHBOARD,
    WidgetNames.SCOPEMEMBERS: WidgetTypes.SCOPEMEMBERS,
}

default_metapanel_widget = dict(
    name=WidgetNames.METAPANELWIDGET,
    config=dict(),
    meta=dict(
        showAll=True,
    ),
)


class ActivityStatus(Enum):
    """The various Activity statuses that are accepted by KE-chain.

    :cvar OPEN: status of activity is open
    :cvar COMPLETED: status of activity is completed
    """

    OPEN = "OPEN"
    COMPLETED = "COMPLETED"


class ScopeStatus(Enum):
    """The various status of a scope.

    .. versionchanged:: 3.0
      The `TEMPLATE` ScopeStatus is deprecated in KE-chain 3

    :cvar ACTIVE: Status of a scope is active (default)
    :cvar CLOSED: Status of a scope is closed
    :cvar TEMPLATE: Status of a scope is a template (not actively used)(deprecated in KE-chain 3.0)
    :cvar DELETING: Status of a scope when the scope is being deleted
    """

    ACTIVE = "ACTIVE"
    CLOSED = "CLOSED"
    TEMPLATE = "TEMPLATE"
    DELETING = "DELETING"


class ScopeCategory(Enum):
    """The various categories of a scope.

    .. versionadded::3.0

    :cvar LIBRARY_SCOPE: The scope is a library scope
    :cvar USER_SCOPE: The scope is a normal user scope
    :cvar TEMPLATE_SCOPE: The scope is a template scope
    """

    LIBRARY_SCOPE = "LIBRARY_SCOPE"
    USER_SCOPE = "USER_SCOPE"
    TEMPLATE_SCOPE = "TEMPLATE_SCOPE"


class ServiceType(Enum):
    """The file types of sim script.

    :cvar PYTHON_SCRIPT: service is a python script
    :cvar NOTEBOOK: service is a jupyter notebook
    """

    PYTHON_SCRIPT = "PYTHON SCRIPT"
    NOTEBOOK = "NOTEBOOK"


class ServiceEnvironmentVersion(Enum):
    """The acceptable versions of python where services run on.

    :cvar PYTHON_3_6: Service execution environment is a python 3.6 container (unsupported)
    :cvar PYTHON_3_7: Service execution environment is a python 3.7 container
    :cvar PYTHON_3_8: Service execution environment is a python 3.8 container
    :cvar PYTHON_3_9: Service execution environment is a python 3.9 container
    :cvar PYTHON_3_10: Service execution environment is a python 3.10 container
    :cvar PYTHON_3_6_NOTEBOOKS: execution environment is a python 3.6 container with jupyter
        notebook preinstalled (unsupported)
    :cvar PYTHON_3_8_NOTEBOOKS: execution environment is a python 3.8 container with jupyter
        notebook preinstalled
    :cvar PYTHON_3_9_NOTEBOOKS: execution environment is a python 3.9 container with jupyter
        notebook preinstalled
    :cvar PYTHON_3_10_NOTEBOOKS: execution environment is a python 3.10 container with jupyter
        notebook preinstalled
    """

    PYTHON_3_6 = "3.6"  # unsupported
    PYTHON_3_7 = "3.7"
    PYTHON_3_8 = "3.8"
    PYTHON_3_9 = "3.9"
    PYTHON_3_10 = "3.10"
    PYTHON_3_6_NOTEBOOKS = "3.6_notebook"  # unsupported
    PYTHON_3_8_NOTEBOOKS = "3.8_notebook"
    PYTHON_3_9_NOTEBOOKS = "3.9_notebook"
    PYTHON_3_10_NOTEBOOKS = "3.10_notebook"


class ServiceScriptUser(Enum):
    """The acceptable usertypes under which a (trusted) service is run.

    :cvar KENODE_USER: Run as "kenode" user. Equivalent to a manager in a scope.
    :cvar TEAMMANAGER_USER: Run as "kenode_team". Equivalent to a manager in a team.
        (disabled until available)
    :cvar CONFIGURATOR_USER: Run as "kenode_configurator". Equivalent to GG:Configurator.
    """

    KENODE_USER = "kenode"
    # TEAMMANAGER_USER = "kenode_team"
    CONFIGURATOR_USER = "kenode_configurator"


class ServiceExecutionStatus(Enum):
    """The acceptable states of a running service.

    :cvar LOADING: Execution is in LOADING state (next RUNNING, FAILED)
    :cvar RUNNING: Execution is in RUNNING state (next COMPLETED, FAILED, TERMINATING)
    :cvar COMPLETED: Execution is in COMPLETED state
    :cvar FAILED: Execution is in FAILED state
    :cvar TERMINATING: Execution is in TERMINATING state (next TERMINATED)
    :cvar TERMINATED: Execution is in TERMINATED state
    """

    LOADING = "LOADING"
    RUNNING = "RUNNING"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    TERMINATING = "TERMINATING"
    TERMINATED = "TERMINATED"


class TeamRoles(Enum):
    """Roles that exist for a team member.

    :cvar MEMBER: A normal team member
    :cvar MANAGER: A team member that may manage the team (add or remove members, change team)
    :cvar OWNER: The owner of a team
    """

    MEMBER = "MEMBER"
    MANAGER = "MANAGER"
    OWNER = "OWNER"


class ScopeRoles(Enum):
    """
    Roles that exist for a member of a scope.

    :cvar MANAGER: owner of the scope, has full rights
    :cvar SUPERVISOR: supervisor member of a scope, has the rights as leadmember and rights to
        manage catalog tasks.
    :cvar LEADMEMBER: elevated member, has assignment rights, no rights on App tasks or
        Catalog tasks.
    :cvar MEMBER: normal member, only has viewing rights
    """

    MANAGER = "manager"
    SUPERVISOR = "supervisor"
    LEADMEMBER = "leadmember"
    MEMBER = "member"


class ScopeMemberActions(Enum):
    """
    Actions to be performed on the members of a scope.

    :cvar ADD: add a member to the scope
    :cvar REMOVE: delete a member from the scope
    """

    ADD = "add"
    REMOVE = "remove"


class ContextType(Enum):
    """Types of Contexts.

    :cvar STATIC_LOCATION: Geolocation / Featurecollection context with a geolocation.
    :cvar TIME_PERIOD: Time Period Context with start_date and due_date
    :cvar TEXT_LABEL: generic textual label
    """

    STATIC_LOCATION = "STATIC_LOCATION"
    TIME_PERIOD = "TIME_PERIOD"
    TEXT_LABEL = "TEXT_LABEL"


class ContextGroup(Enum):
    """
    Context may have a context_group.

    ..versionadded: 3.11

    This is for context API versions 1.2.0 or later.

    :cvar UNDEFINED: UNDEFINED
    :cvar DISCIPLINE: Discipline, in nl: Discipline
    :cvar ASSET: Asset, in nl: Object, Kunstwerk
    :cvar DEPARTMENT: Department, in nl: Onderdeel, Afdeling
    :cvar PERIOD: Workperiod, in nl: Werkperiode
    :cvar LOCATION: Location, in nl: Locatie
    :cvar PHASE: Phase, in nl: Fase
    :cvar REQUIREMENT: Requirement, in nl: Eis
    :cvar EXTERNALID: External identifier, to be used to provide a generic link to an
        external application
    :cvar WORKPACKAGE: Workpackage, in nl: Werkpakket
    """

    UNDEFINED = "UNDEFINED"
    DISCIPLINE = "DISCIPLINE"  # nl: Discipline
    ASSET = "ASSET"  # nl: Object, Kunstwerk
    DEPARTMENT = "DEPARTMENT"  # nl: Onderdeel, Afdeling
    PERIOD = "WORKPERIOD"  # nl: Werkperiode
    LOCATION = "LOCATION"  # nl: Locatie
    PHASE = "PHASE"  # nl: Fase
    REQUIREMENT = "REQUIREMENT"  # nl: Eis
    EXTERNALID = "EXTERNALID"
    WORKPACKAGE = "WORKPACKAGE"  # nl: Werkpakket


class KechainEnv(Enum):
    """Environment variables that can be set for pykechain.

    :cvar KECHAIN_URL: full url of KE-chain where to connect to eg: 'https://<some>.ke-chain.com'
    :cvar KECHAIN_TOKEN: authentication token for the KE-chain user provided from KE-chain user
        account control
    :cvar KECHAIN_USERNAME: the username for the credentials
    :cvar KECHAIN_PASSWORD: the password for the credentials
    :cvar KECHAIN_SCOPE: the name of the project / scope. Should be unique, otherwise use scope_id
    :cvar KECHAIN_SCOPE_ID: the UUID of the project / scope.
    :cvar KECHAIN_FORCE_ENV_USE: set to 'true', '1', 'ok', or 'yes' to always use the environment
        variables.
    :cvar KECHAIN_SCOPE_STATUS: the status of the Scope to retrieve, defaults to None to retrieve
        all scopes
    :cvar KECHAIN_CHECK_CERTIFICATES: if the certificates of the URL should be checked.
    """

    KECHAIN_FORCE_ENV_USE = "KECHAIN_FORCE_ENV_USE"
    KECHAIN_URL = "KECHAIN_URL"
    KECHAIN_TOKEN = "KECHAIN_TOKEN"
    KECHAIN_USERNAME = "KECHAIN_USERNAME"
    KECHAIN_PASSWORD = "KECHAIN_PASSWORD"
    KECHAIN_SCOPE = "KECHAIN_SCOPE"
    KECHAIN_SCOPE_ID = "KECHAIN_SCOPE_ID"
    KECHAIN_SCOPE_STATUS = "KECHAIN_SCOPE_STATUS"
    KECHAIN_CHECK_CERTIFICATES = "KECHAIN_CHECK_CERTIFICATES"


class SortTable(Enum):
    """The acceptable sorting options for a grid/table.

    :cvar ASCENDING: Table is sorted in ASCENDING ORDER
    :cvar DESCENDING: Table is sorted in DESCENDING ORDER
    """

    ASCENDING = "ASC"
    DESCENDING = "DESC"


class Alignment(Enum):
    """The alignment options for attachment viewer, navigation bar widgets and service widgets.

    :cvar LEFT: Aligned to the left
    :cvar CENTER: Aligned to the center
    :cvar RIGHT: Aligned to the right
    """

    LEFT = "left"
    CENTER = "center"
    RIGHT = "right"


class NavigationBarAlignment(Alignment):
    """The acceptable alignment options for a Navigation Bar Widget."""

    pass


class SidebarItemAlignment(Enum):
    """The acceptable alignment options for sidebar button.

    :cvar TOP: "top"
    :cvar BOTTOM: "bottom"
    """

    TOP = "top"
    BOTTOM = "bottom"


class SidebarButtonAlignment(SidebarItemAlignment):
    """Compatibility enumeration class for the SidebarItemAlignment."""

    pass


class SidebarType(Enum):
    """The types that can exist as a Sidebar Item.

    :cvar BUTTON: a button,
    :cvar CARD: a card
    """

    BUTTON = "BUTTON"
    CARD = "CARD"


class PaperSize(Enum):
    """The acceptable paper sizes options for a downloaded PDF.

    :cvar A0: Paper of size A0
    :cvar A1: Paper of size A1
    :cvar A2: Paper of size A2
    :cvar A3: Paper of size A3
    :cvar A4: Paper of size A4
    """

    A0 = "a0paper"
    A1 = "a1paper"
    A2 = "a2paper"
    A3 = "a3paper"
    A4 = "a4paper"
    AUTO = "automatic"


class PaperOrientation(Enum):
    """The acceptable paper orientation options for a downloaded PDF.

    :cvar PORTRAIT: Paper of orientation 'portrait'
    :cvar LANDSCAPE: Paper of orientation 'landscape'
    """

    PORTRAIT = "portrait"
    LANDSCAPE = "landscape"


class PropertyVTypes(Enum):
    """The VTypes (or validator types) that are allowed in the json.

    This corresponds to the various validator classes which SHOULD be named:
       `vtype[0].upper() + vtype[1:]`
       eg: 'numbericRangeValidator' has an implementation class of 'NumericRangeValidator'

    .. versionadded:: 2.2

    :cvar NONEVALIDATOR: noneValidator - No validation is done
    :cvar NUMERICRANGE: numericRangeValidator
    :cvar BOOLEANFIELD: booleanFieldValidator
    :cvar REQUIREDFIELD: requiredFieldValidator
    :cvar EVENNUMBER: evenNumberValidator
    :cvar ODDNUMBER: oddNumberValidator
    :cvar REGEXSTRING: regexStringValidator
    :cvar SINGLEREFERENCE: 'singleReferenceValidator'
    :cvar FILEEXTENSION: 'fileExtensionValidator'
    :cvar FILESIZE: 'fileSizeValidator'
    """

    NONEVALIDATOR = "noneValidator"
    NUMERICRANGE = "numericRangeValidator"
    BOOLEANFIELD = "booleanFieldValidator"
    REQUIREDFIELD = "requiredFieldValidator"
    EVENNUMBER = "evenNumberValidator"
    ODDNUMBER = "oddNumberValidator"
    REGEXSTRING = "regexStringValidator"
    SINGLEREFERENCE = "singleReferenceValidator"
    FILEEXTENSION = "fileExtensionValidator"
    FILESIZE = "fileSizeValidator"

    # fallback
    ALWAYSALLOW = "alwaysAllowValidator"


class ValidatorEffectTypes(Enum):
    """The effects that can be attached to a validator.

    .. versionadded:: 2.2

    :cvar NONE_EFFECT: noneEffect
    :cvar VISUALEFFECT: visualEffect
    :cvar TEXT_EFFECT: textEffect
    :cvar ERRORTEXT_EFFECT: errorTextEffect
    :cvar HELPTEXT_EFFECT: helpTextEffect
    """

    NONE_EFFECT = "noneEffect"
    VISUALEFFECT = "visualEffect"
    TEXT_EFFECT = "textEffect"
    ERRORTEXT_EFFECT = "errorTextEffect"
    HELPTEXT_EFFECT = "helpTextEffect"


class PropertyRepresentation(Enum):
    """
    The Representation configuration to display a property value.

    .. versionadded:: 3.0
    .. versionchanged:: 3.11 added geocoordinate in line with KE-chain v2021.5.0

    :cvar DECIMAL_PLACES: Amount of decimal places to show the number
    :cvar SIGNIFICANT_DIGITS: Number (count) of significant digits to display the number
    :cvar LINK_TARGET: configuration of a link to open the link in a new browsertab or not.
    :cvar BUTTON: options to represent the choices of a select-list
    :cvar THOUSANDS_SEPARATOR: option to display the thousand separator
    :cvar AUTOFILL: option to autofill the content of the property
    :cvar GEOCOORDINATE: option to display an alternative representation for the geocoordinate
    :cvar USE_PROPERTY_NAME: option to display the name of a property for a part actvity ref prop
    """

    DECIMAL_PLACES = "decimalPlaces"
    SIGNIFICANT_DIGITS = "significantDigits"
    LINK_TARGET = "linkTarget"
    BUTTON = "buttonRepresentation"
    THOUSANDS_SEPARATOR = "thousandsSeparator"
    AUTOFILL = "autofill"
    GEOCOORDINATE = "geoCoordinate"
    USE_PROPERTY_NAME = "usePropertyName"
    CAMERA_SCANNER_INPUT = "cameraScannerInput"


class GeoCoordinateConfig(Enum):
    """GeoCoordinate Configuration Enumerations.

    :cvar APPROX_ADDRESS: represent the coordinate as approximate address (lookup by Google)
    :cvar RD_AMERSFOORT: represent the coordinate as Amersfoort / RN New (epsg: 28992)
    :cvar DD: represent the coordinate as Decimal Degrees (WGS84, epsg:4326)
    :cvar DMS: represent the coordinate as as Degrees Minutes Seconds (WGS84, epsg:4326)

    """

    APPROX_ADDRESS = "approx_address"  # As approximated address
    # Amersfoort/RD (epsg: 28992) https://www.spatialreference.org/ref/epsg/amersfoort-rd-new/
    RD_AMERSFOORT = "rd_amersfoort"
    DD = (
        # As WSG84 (epsg:4326) decimal degrees representation first lat (-90,+90) then lng
        # (-180,+180)
        "dd"
    )
    DMS = (
        # As WSG84 (epsg:4326) degrees, minutes, seconds representation first lat N/S then lng E/W
        "dms"
    )


class OtherRepresentations(Enum):
    """
    Other representations used in KE-chain.

    :cvar CUSTOM_ICON: different font-awesome icons
    """

    CUSTOM_ICON = "customIcon"


class _AllRepresentations(PropertyRepresentation, OtherRepresentations):
    pass


class ShowColumnTypes(Enum):
    """The columns that can be shown in a Property grid.

    .. versionadded:: 2.3

    :cvar UNIT: unit
    :cvar DESCRIPTION: description
    """

    UNIT = "unit"
    DESCRIPTION = "description"


class ScopeWidgetColumnTypes(Enum):
    """The columns that can be shown in a Scope widget grid.

    .. versionadded:: 3.0

    :cvar PROJECT_NAME: Name
    :cvar START_DATE: Start date
    :cvar DUE_DATE: Due date
    :cvar PROGRESS: Progress
    :cvar STATUS: Status
    :cvar TAGS: Tags
    """

    PROJECT_NAME = "Name"
    START_DATE = "Start date"
    DUE_DATE = "Due date"
    PROGRESS = "Progress"
    STATUS = "Status"
    TAGS = "Tags"


class FilterType(Enum):
    """The type of pre-filters that can be set on a Multi Reference Property.

    .. versionadded:: 3.0

    :cvar GREATER_THAN_EQUAL: 'gte'
    :cvar LOWER_THAN_EQUAL: 'lte'
    :cvar CONTAINS: 'icontains'
    :cvar EXACT: 'exact'
    """

    GREATER_THAN_EQUAL = "gte"
    LOWER_THAN_EQUAL = "lte"
    CONTAINS = "icontains"
    CONTAINS_SET = "contains"
    EXACT = "exact"


class ProgressBarColors(Enum):
    """
    Some basic colors that can be set on a Progress Bar inside a Progress Bar Widget.

    .. versionadded:: 3.0

    :cvar BLACK: '#000000'
    :cvar WHITE: '#FFFFFF'
    :cvar RED: 'FF0000'
    :cvar LIME: '#00FF00'
    :cvar BLUE: '#0000FF'
    :cvar YELLOW: '#FFFF00'
    :cvar CYAN: '#00FFFF'
    :cvar MAGENTA: '#FF00FF'
    :cvar SILVER: '#C0C0C0'
    :cvar GRAY: '#808080'
    :cvar MAROON: '#800000'
    :cvar OLIVE: '#808000'
    :cvar GREEN: '#008000'
    :cvar PURPLE: '#800080'
    :cvar TEAL: '#008080'
    :cvar NAVY: '#000080'
    :cvar DEFAULT_COMPLETED: '#339447'
    :cvar DEFAULT_IN_PROGRESS: '#FF6600'
    :cvar DEFAULT_NO_PROGRESS: '#EEEEEE'
    :cvar DEFAULT_IN_PROGRESS_BACKGROUND: '#FC7C3D'
    """

    BLACK = "#000000"
    WHITE = "#FFFFFF"
    RED = "#FF0000"
    LIME = "#00FF00"
    BLUE = "#0000FF"
    YELLOW = "#FFFF00"
    CYAN = "#00FFFF"
    MAGENTA = "#FF00FF"
    SILVER = "#C0C0C0"
    GRAY = "#808080"
    MAROON = "#800000"
    OLIVE = "#808000"
    GREEN = "#008000"
    PURPLE = "#800080"
    TEAL = "#008080"
    NAVY = "#000080"
    DEFAULT_COMPLETED = "#339447"
    DEFAULT_IN_PROGRESS = "#FF6600"
    DEFAULT_NO_PROGRESS = "#EEEEEE"
    DEFAULT_IN_PROGRESS_BACKGROUND = "#FC7C3D"


class LinkTargets(Enum):
    """
    Target for the CardWidget link and Link property representations.

    .. versionadded:: 3.0

    :cvar SAME_TAB: "_self"
    :cvar NEW_TAB: "_blank"
    """

    SAME_TAB = "_self"
    NEW_TAB = "_blank"


class CardWidgetLinkTarget(LinkTargets):
    """Target for the CardWidget, remaining for backwards compatibility."""

    pass


class CardWidgetLinkValue(Enum):
    """
    Link Value for the CardWidget.

    .. versionadded:: 3.0

    :cvar EXTERNAL_LINK: "External link"
    :cvar TASK_LINK: "Task link"
    :cvar NO_LINK: "No link"
    """

    EXTERNAL_LINK = "External link"
    TASK_LINK = "Task link"
    TREE_VIEW = "Tree view"
    NO_LINK = "No link"


class CardWidgetImageValue(Enum):
    """
    Image for the CardWidget.

    .. versionadded:: 3.0

    :cvar CUSTOM_IMAGE: "Custom image"
    :cvar NO_IMAGE: "No image"
    """

    CUSTOM_IMAGE = "Custom image"
    NO_IMAGE = "No image"


class KEChainPages(Enum):
    """
    URL names of built-in KE-chain pages.

    :cvar DETAIL: "detail"
    :cvar FORMS: "forms"
    :cvar TASKS: "activities"
    :cvar WORK_BREAKDOWN: "activitytree"
    :cvar CATALOG_FORMS: "catalogforms"
    :cvar CONTEXTS: "contexts"
    :cvar WORKFLOWS: "workflows"
    :cvar DATA_MODEL: "productmodel"
    :cvar EXPLORER: "product"
    :cvar SERVICES: "scripts"
    """

    DETAIL = "detail"
    FORMS = "forms"
    TASKS = "activities"
    WORK_BREAKDOWN = "activitytree"
    CATALOG_FORMS = "catalogforms"
    CONTEXTS = "contexts"
    WORKFLOWS = "workflows"
    DATA_MODEL = "productmodel"
    EXPLORER = "product"
    SERVICES = "scripts"
    CATALOG_WBS = "catalogtree"
    APP_WBS = "apptree"


KEChainPageLabels = {
    KEChainPages.DETAIL: "Project details",
    KEChainPages.FORMS: "Forms",
    KEChainPages.TASKS: "Tasks",
    KEChainPages.WORK_BREAKDOWN: "Work Breakdown",
    KEChainPages.CATALOG_FORMS: "Template forms",
    KEChainPages.CONTEXTS: "Contexts",
    KEChainPages.WORKFLOWS: "Workflows",
    KEChainPages.CATALOG_WBS: "Catalog",
    KEChainPages.APP_WBS: "App Screens",
    KEChainPages.DATA_MODEL: "Data model",
    KEChainPages.EXPLORER: "Explorer",
    KEChainPages.SERVICES: "Scripts",
}

KEChainPageLabels_nl = {
    KEChainPages.DETAIL: "Project details",
    KEChainPages.FORMS: "Formulieren",
    KEChainPages.TASKS: "Taken",
    KEChainPages.WORK_BREAKDOWN: "Taakverdeling",
    KEChainPages.CATALOG_FORMS: "Sjablonen",
    KEChainPages.CONTEXTS: "Contexten",
    KEChainPages.WORKFLOWS: "Workflows",
    KEChainPages.CATALOG_WBS: "Catalogus",
    KEChainPages.APP_WBS: "App schermen",
    KEChainPages.DATA_MODEL: "Data model",
    KEChainPages.EXPLORER: "Explorer",
    KEChainPages.SERVICES: "Scripts",
}

CardWidgetKEChainPageLink = {
    KEChainPages.DETAIL: "Project",
    KEChainPages.FORMS: "Forms",
    KEChainPages.TASKS: "Tasks",
    KEChainPages.DATA_MODEL: "Model",
    KEChainPages.EXPLORER: "Explorer",
    KEChainPages.SERVICES: "Script",
    KEChainPages.WORK_BREAKDOWN: "Work Breakdown",
    KEChainPages.CATALOG_WBS: "Catalog Tasks",
    KEChainPages.APP_WBS: "App Tasks",
    KEChainPages.CATALOG_FORMS: "Template forms",
    KEChainPages.CONTEXTS: "Contexts",
    KEChainPages.WORKFLOWS: "Workflows",
}

KEChainPageIcons = {
    KEChainPages.DETAIL: "bookmark",
    KEChainPages.FORMS: "file-contract",
    KEChainPages.TASKS: "edit",
    KEChainPages.WORK_BREAKDOWN: "sitemap",
    KEChainPages.CATALOG_FORMS: "file-export",
    KEChainPages.CONTEXTS: "tags",
    KEChainPages.WORKFLOWS: "directions",
    KEChainPages.CATALOG_WBS: "books",
    KEChainPages.APP_WBS: "tablet-alt",
    KEChainPages.DATA_MODEL: "cube",
    KEChainPages.EXPLORER: "folder",
    KEChainPages.SERVICES: "file-code",
}


class SubprocessDisplayMode(Enum):
    """
    URL variations to vary the display of a subprocess activity.

    :cvar ACTIVITIES: "activities"
    :cvar TREEVIEW: "treeview"
    """

    ACTIVITIES = "activities"
    TREEVIEW = "treeview"


class URITarget(Enum):
    """
    Side-bar button redirect options.

    :cvar INTERNAL: "internal"
    :cvar EXTERNAL: "external"
    :cvar NEW: "_new"
    :cvar SELF: "_self"
    :cvar BLANK: "_blank"
    :cvar PARENT: "_parent"
    :cvar TOP: "_top"
    """

    INTERNAL = "internal"
    EXTERNAL = "external"
    NEW = "_new"
    SELF = "_self"
    BLANK = "_blank"
    PARENT = "_parent"
    TOP = "_top"


class FontAwesomeMode(Enum):
    """
    Options to display the same icon.

    Source:
    https://fontawesome.com/how-to-use/on-the-web/setup/getting-started

    :cvar SOLID: "solid"
    :cvar REGULAR: "regular"
    :cvar LIGHT: "light"
    """

    SOLID = "solid"
    REGULAR = "regular"
    LIGHT = "light"


class SidebarAccessLevelOptions(Enum):
    """
    Options for access level options for the sidebar.

    :cvar IS_MEMBER: "is_member"
    :cvar IS_LEAD_MEMBER: "is_leadmember"
    :cvar IS_SUPERVISOR: "is_supervisor"
    :cvar IS_MANAGER: "is_manager"
    """

    IS_MEMBER = "is_member"
    IS_LEAD_MEMBER = "is_leadmember"
    IS_SUPERVISOR = "is_supervisor"
    IS_MANAGER = "is_manager"


class MinimumAccessLevelOptions(SidebarAccessLevelOptions):
    """Options for minumum access level options for the sidebar."""

    pass


class MaximumAccessLevelOptions(SidebarAccessLevelOptions):
    """Options for maximum access level options for the sidebar."""

    pass


class SelectListRepresentations(Enum):
    """
    Options in which a single-select list property options are displayed.

    :cvar DROP_DOWN: "dropdown"
    :cvar CHECK_BOXES: "checkboxes"
    :cvar BUTTONS: "buttons"
    """

    DROP_DOWN = "dropdown"
    CHECK_BOXES = "checkboxes"
    BUTTONS = "buttons"


class ImageFitValue(Enum):
    """
    Options to fit an image on a CardWidget or AttachmentViewerWidget.

    This is a subset from the `object-fit property`_ in HTML.

    :cvar CONTAIN: scale the image to fit within the widget
    :cvar COVER: scale the image to cover the entire widget

    .. _object-fit property: https://developer.mozilla.org/en-US/docs/Web/CSS/object-fit
    """

    CONTAIN = "contain"
    COVER = "cover"


class WidgetTitleValue(Enum):
    """
    Options to configure the title of a widget.

    :cvar DEFAULT: Use the default title of the widget type.
    :cvar NO_TITLE: Show no title.
    :cvar CUSTOM_TITLE: Show a custom title text.
    """

    DEFAULT = "Default"
    NO_TITLE = "No title"
    CUSTOM_TITLE = "Custom title"


class NotificationStatus(Enum):
    """
    Options to retrieve a Notification based on its status.

    normal lifecycle:
    - DRAFT, when a message is first saved to the backend and the status is still in draft.
        next states: READY
    - READY: when the message is ready for processing, it is complete and is to be processed
        next states: PROCESSING
    - PROCESSING: when the message is in the process of being send out
        next states: COMPLETED, FAILED
    - COMPLETED: when the message is successfully sent out
        next states: ARCHIVED
    - FAILED: when the message is not successfully sent out
        next states: ARCHIVED
    - ARCHIVED: when the message is archives and waiting for its deletion against a certain
        retention policy
        next states: None

    :cvar ARCHIVED: "archived" notifications
    :cvar COMPLETED: "completed" notifications
    :cvar DRAFT: "draft" notifications
    :cvar FAILED: "failed" notifications
    :cvar PROCESSING: "processing" notifications
    :cvar READY: "ready" notifications
    """

    ARCHIVED = "ARCHIVED"
    COMPLETED = "COMPLETED"
    DRAFT = "DRAFT"
    FAILED = "FAILED"
    PROCESSING = "PROCESSING"
    READY = "READY"


class NotificationEvent(Enum):
    """
    Options to retrieve a Notification based on its event.

    :cvar SHARE_ACTIVITY_LINK: notifications generated by sharing the link of an `Activity`
    :cvar EXPORT_ACTIVITY_ASYNC: notifications generated by exporting an `Activity`
    :cvar SHARE_ACTIVITY_PDF: notifications generated by sharing the pdf of an `Activity`
    """

    SHARE_ACTIVITY_LINK = "SHARE_ACTIVITY_LINK"
    EXPORT_ACTIVITY_ASYNC = "EXPORT_ACTIVITY_ASYNC"
    SHARE_ACTIVITY_PDF = "SHARE_ACTIVITY_PDF"


class NotificationChannels(Enum):
    """
    Options to retrieve a Notification based on its channel.

    :cvar EMAIL: email notification
    :cvar APP: app notification
    """

    EMAIL = "EMAIL"
    APP = "APP"


class LanguageCodes(Enum):
    """
    Options for the language setting of a user.

    :cvar ENGLISH: English
    :cvar FRENCH: French
    :cvar GERMAN: German
    :cvar DUTCH: Dutch
    :cvar ITALIAN: Italian
    """

    ENGLISH = "en"
    FRENCH = "fr"
    GERMAN = "de"
    DUTCH = "nl"
    ITALIAN = "it"


class ImageSize(Enum):
    """
    Options for the Image Size the picture would be saved as from an Attachment Property.

    :cvar SQXS: SQXS (100, 100)  # pixels square
    :cvar XS: XS (100, )  # pixels width
    :cvar S: S (320, )
    :cvar SQS: SQS (320, 320)
    :cvar M: M (640, )
    :cvar SQM: SQM (640, 640)
    :cvar L: L (1024, )
    :cvar SQL: SQL (1024, 1024)
    :cvar XL: XL (2048, )
    :cvar SQXL: SQXL (2048, 2048)
    :cvar XXL: XXL (4096, )
    :cvar SQXXL: SQXXL (4096, 4096)
    """

    SQXS = "SQXS"
    XS = "XS"
    S = "S"
    SQS = "SQS"
    M = "M"
    SQM = "SQM"
    L = "L"
    SQL = "SQL"
    XL = "XL"
    SQXL = "SQXL"
    XXL = "XXL"
    SQXXL = "SQXXL"


class FormCategory(Enum):
    """
    Options for the Category of a Form.

    :cvar MODEL: Model
    :cvar INSTANCE: Instance
    """

    MODEL = "MODEL"
    INSTANCE = "INSTANCE"


class WorkflowCategory(Enum):
    """
    Options for the Category of a Workflow.

    :cvar CATALOG: Catalog Workflow (immutable)
    :cvar DEFINED: Defined Workflow belonging to a scope
    """

    CATALOG = "CATALOG"
    DEFINED = "DEFINED"


class TransitionType(Enum):
    """
    Options for the Type of a Transition.

    :cvar INITIAL: Initial transition, the initial transition to follow when the form is created.
    :cvar GLOBAL: Global transition, possibility to transition to all statuses from any
    :cvar DIRECTED: A Directed transition, a transition with a specific from -> to direction
    """

    INITIAL = "INITIAL"
    GLOBAL = "GLOBAL"
    DIRECTED = "DIRECTED"


class StatusCategory(Enum):
    """
    Category of statuses.

    :cvar UNDEFINED: Undefined status
    :cvar TODO: Todo status
    :cvar INPROGRESS: In progress status
    :cvar DONE: Done status
    """

    UNDEFINED = "UNDEFINED"
    TODO = "TODO"
    INPROGRESS = "INPROGRESS"
    DONE = "DONE"
